from dataclasses import dataclass


@dataclass
class FcsConfig:
    def __init__(self, destination: str, nexus_data_name: str = "data") -> None:
        """Initialises the FcsConfig class.

        Args:
            destination: The destination directory for the output data.
            nexus_data_name: The name of the data directory in the Nexus model.
        """
        self.output_dir = destination
        self.use_reservoir_names = True
        self.abs_paths_to_keep: list[str] = []
        self.use_one_dir = True
        self.input_data_dir = ''
        self.path_components_to_keep = None
        self.locals_to_data = True
        self.output_data_dir = nexus_data_name
        self.dry_run = False
        self.compress = False
        self.output_log_file = False



from __future__ import annotations

import warnings
from dataclasses import dataclass, field
from enum import Enum
import re
from typing import Optional, Union
import numpy as np
import pandas as pd
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_ARRAY_KEYWORDS, HYD_TABLE_HEADER_COLS, HYD_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_PRESSURE_KEYWORDS, HYD_SINGLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_KEYWORDS_VALUE_FLOAT, HYD_WATINJ_KEYWORDS_VALUE_FLOAT
from ResSimpy.Nexus.NexusKeywords.hyd_keywords import HYD_ALQ_KEYWORD, HYD_ALQ_OPTIONS
from ResSimpy.Enums.UnitsEnum import UnitSystem, SUnits, TemperatureUnits
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import HydraulicsUnits

from ResSimpy.Utils.factory_methods import get_empty_dict_union
import ResSimpy.Nexus.nexus_file_operations as nfo


@dataclass(kw_only=True, repr=False)  # Doesn't need to write an _init_, _eq_ methods, etc.
class NexusHydraulicsMethod(DynamicProperty):
    """Class to hold Nexus Hydraulics properties.

    Attributes:
        file (NexusFile): Nexus hydraulics properties file object
        input_number (int): Hydraulics properties method number in Nexus fcs file
        properties (dict[str, Union[str, int, float, Enum, list[str], pd.DataFrame,
                    dict[str, Union[float, pd.DataFrame]]]]):
            Dictionary holding all properties for a specific hydraulics properties method. Defaults to empty dictionary.
    """

    # General parameters
    file: Optional[NexusFile]
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                     dict[str, Union[float, pd.DataFrame]]]] = \
        field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    ratio_thousands: bool

    def __init__(self, file: Optional[NexusFile], input_number: int, model_unit_system: UnitSystem,
                 ratio_thousands: bool = True,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                      dict[str, Union[float, pd.DataFrame]]]]] = None) -> None:
        """Initialises the NexusHydraulicsMethod class.

        Args:
            file (NexusFile): NexusFile object associated with the hydraulics method.
            input_number (int): method number for the hydraulics method.
            model_unit_system (UnitSystem): unit system used in the model.
            ratio_thousands (bool): ??
            properties: dictionary of key properties for the hydraulics method. Defaults to None.
        """
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        self.unit_system = model_unit_system
        self.ratio_thousands = ratio_thousands
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'QOIL': ('surface_oil_rate', float),
            'QLIQ': ('surface_liquid_rate', float),
            'QWAT': ('surface_water_rate', float),
            'QGAS': ('surface_gas_rate', float),
            'QWGAS': ('surface_wet_gas_rate', float),
            'MMW': ('mean_molecular_weight', float),
            'WCUT': ('watercut', float),
            'OCUT': ('oilcut', float),
            'PIN': ('inlet_pressure', float),
            'POUT': ('outlet_pressure', float),
            'BHP': ('bottomhole_pressure', float),
            'THP': ('tubinghead_pressure', float),
            'GLR': ('gas_liquid_ratio', float),
            'GOR': ('gas_oil_ratio', float),
            'OGR': ('oil_gas_ratio', float),
            'WGR': ('water_gas_ratio', float),
            'GWR': ('gas_water_ratio', float),
            'WWGR': ('water_wet_gas_ratio', float),
            'LENGTH': ('length', float),
            'DATUM': ('datum_depth', float),
            'DZ': ('depth_change', float),
            'DZW': ('hydraulic_table_vertical_distance', float),
            'GRAD': ('injected_fluid_pressure_gradient', float),
            'VISC': ('viscosity', float),
            'DIAM': ('diameter', float),
            'ROUGHNESS': ('roughness', float),
        }
        return keywords

    @property
    def units(self) -> HydraulicsUnits:
        """Returns the attribute to unit map for the Hydraulics method."""
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return HydraulicsUnits(unit_system=self.unit_system, ratio_thousands=self.ratio_thousands)

    def to_string(self) -> str:
        """Create string with hydraulics data in Nexus file format."""
        printable_str = ''
        hyd_dict = self.properties
        for key, value in hyd_dict.items():
            if isinstance(value, pd.DataFrame):
                if key not in ['HYD_TABLE']:
                    printable_str += f'{key}\n'
                printable_str += value.to_string(na_rep='', index=False) + '\n'
                if key == 'LIMITS':
                    printable_str += 'ENDLIMITS\n'
                printable_str += '\n'
            elif key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
                elif isinstance(value, SUnits):
                    printable_str += f'SUNITS {value.value}\n'
            elif key == 'ALQ':
                printable_str += f'{key}'
                if 'ALQ_PARAM' in hyd_dict.keys():
                    printable_str += f" {hyd_dict['ALQ_PARAM']}"
                if isinstance(value, np.ndarray):
                    printable_str += f" {' '.join([str(val) for val in value])}\n"
                else:
                    printable_str += f' {value}\n'
            elif key == 'WATINJ' and isinstance(value, dict):
                printable_str += f'{key}\n'
                for watinj_key in value.keys():
                    printable_str += f'    {watinj_key} {value[watinj_key]}\n'
            elif key not in ['ALQ', 'ALQ_PARAM', 'WATINJ']:
                if isinstance(value, np.ndarray):
                    printable_str += f"{key} {' '.join([str(val) for val in value])}\n"
                elif value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        printable_str += '\n'
        return printable_str

    def read_properties(self) -> None:
        """Read Nexus hydraulics file contents and populate the NexusHydraulicsMethod object."""
        if self.file is None:
            warnings.warn('No file provided to read Nexus hydraulics properties from.')
            return
        original_file_as_list = self.file.get_flat_list_str_file

        # Loop through file_as_list and fix any line continuations that use '>' at the end of the line
        file_as_list: list[str] = []
        if len(original_file_as_list) > 0:
            file_as_list = [original_file_as_list[0]]
            line_indx = 1
            while line_indx < len(original_file_as_list):
                prev_line = file_as_list[-1]
                line = original_file_as_list[line_indx]
                if prev_line.rstrip().endswith(('>', '>\n')):
                    file_as_list[-1] = prev_line.split('>')[0] + ' ' + line.strip()
                else:
                    file_as_list.append(line)
                line_indx += 1

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize flags and containers to use to record properties as we iterate through aquifer file contents
        # Dictionary to record start and ending indices for tables
        hyd_table_indices: dict[str, list[int]] = {}
        # Dictionary of flags indicating which tables are being read
        table_being_read: dict[str, bool] = {'HYD_TABLE': False, 'LIMITS': False}
        start_reading_table = False
        # Containers for handling WATERINJ data
        watinj_dict: dict[str, float] = {}
        found_waterinj = False

        line_indx = 0
        for line in file_as_list:

            # Find arrays of parameters, e.g., QOIL 1.0 10. 100., or GOR 0.0 0.5 1.0
            potential_keyword = nfo.check_list_tokens(HYD_ARRAY_KEYWORDS + HYD_ALQ_KEYWORD, line)
            if potential_keyword is not None and not table_being_read['LIMITS']:
                line_elems = line.split('!')[0].split()
                next_val = nfo.get_expected_token_value(potential_keyword, line, file_as_list)
                if potential_keyword == 'ALQ' and next_val in HYD_ALQ_OPTIONS:
                    self.properties['ALQ_PARAM'] = next_val
                    keyword_index = line_elems.index(next_val)
                else:
                    keyword_index = line_elems.index(potential_keyword)
                self.properties[potential_keyword] = np.fromstring(' '.join(line_elems[keyword_index+1:]), sep=' ')

            # Create WATINJ property if needed
            if nfo.check_token('WATINJ', line):
                found_waterinj = True

            # Handle DATGRAD property
            if nfo.check_token('DATGRAD', line):
                self.properties['DATGRAD'] = nfo.get_expected_token_value('DATGRAD', line, file_as_list)

            # Find HYD key-value pairs, such as LENGTH 3000, DATUM 7000 or DATGRAD GRAD
            potential_keyword = nfo.check_list_tokens(HYD_KEYWORDS_VALUE_FLOAT + HYD_WATINJ_KEYWORDS_VALUE_FLOAT, line)
            if potential_keyword is not None:
                if found_waterinj and potential_keyword in HYD_WATINJ_KEYWORDS_VALUE_FLOAT:
                    watinj_dict[potential_keyword] = float(
                        nfo.get_expected_token_value(potential_keyword, line, file_as_list))
                elif potential_keyword in HYD_KEYWORDS_VALUE_FLOAT:
                    self.properties[potential_keyword] = float(
                        nfo.get_expected_token_value(potential_keyword, line, file_as_list))

            # Find standalone hydraulics keywords
            potential_keyword = nfo.check_list_tokens(HYD_SINGLE_KEYWORDS, line)
            if potential_keyword is not None:
                self.properties[potential_keyword] = ''

            # Find starting and ending indices for hydraulic limits table
            if nfo.check_token('LIMITS', line):
                hyd_table_indices['LIMITS'] = [line_indx+1, len(file_as_list)]
                table_being_read['LIMITS'] = True
                start_reading_table = True
            if nfo.check_token('ENDLIMITS', line):
                hyd_table_indices['LIMITS'][1] = line_indx
                table_being_read['LIMITS'] = False
                start_reading_table = False

            # Find ending index of hydraulics table
            if start_reading_table and table_being_read['HYD_TABLE']:
                for potential_endkeyword in HYD_KEYWORDS:
                    if nfo.check_token(potential_endkeyword, line):
                        hyd_table_indices['HYD_TABLE'][1] = line_indx
                        start_reading_table = False
                        table_being_read['HYD_TABLE'] = False
                        break
            # Find starting index of hydraulics table
            for header_keyword in HYD_TABLE_HEADER_COLS:
                if nfo.check_token(header_keyword, line):
                    hyd_table_indices['HYD_TABLE'] = [line_indx, len(file_as_list)]
                    table_being_read['HYD_TABLE'] = True
                    start_reading_table = True

            line_indx += 1

        # Read in hydraulics tables
        for key in hyd_table_indices.keys():
            if key == 'HYD_TABLE':
                # Get hydraulic table data
                hyd_df: pd.DataFrame = nfo.read_table_to_df(file_as_list[
                    hyd_table_indices[key][0]+1:hyd_table_indices[key][1]], noheader=True)
                # Resolve hydraulic table header
                hyd_df_header_names = [header_name for header_name in file_as_list[
                    hyd_table_indices[key][0]].split('!')[0].split() if not re.search('BHP', header_name)]
                pressure_keyword_map = {'PIN': 'POUT', 'POUT': 'PIN', 'THP': 'BHP'}
                hyd_table_pressure_key = ''  # Determines what type of pressures are in table, i.e., PIN, POUT or BHP
                for pkey in HYD_PRESSURE_KEYWORDS:
                    if pkey in self.properties.keys():
                        if pkey == 'THP':
                            # For VIP compatibility, indicates units
                            # for GOR/GLR are SCF/STB rather than MSCF/STB
                            self.ratio_thousands = False
                        hyd_table_pressure_key = pressure_keyword_map[pkey]
                        break
                hyd_df_header_names = hyd_df_header_names + \
                    [hyd_table_pressure_key+str(i) for i in range(hyd_df.shape[1] - len(hyd_df_header_names))]
                hyd_df.columns = hyd_df_header_names
                # Set appropriate hydraulic table property
                self.properties[key] = hyd_df
            else:
                self.properties[key] = nfo.read_table_to_df(file_as_list[
                    hyd_table_indices[key][0]:hyd_table_indices[key][1]])

        # Correctly set WATINJ properties, if needed
        if found_waterinj:
            self.properties['WATINJ'] = watinj_dict

    def __hash__(self) -> int:
        """Returns the hash of the object. Excludes the file and input number attributes."""
        property_hash = self.convert_to_hashable(self.properties)
        return hash((property_hash, self.unit_system, self.ratio_thousands))



"""Data structure for holding Nexus options data."""
from dataclasses import dataclass, field
from typing import Optional, Union
import numpy as np
import pandas as pd
from enum import Enum
from ResSimpy.Enums.UnitsEnum import TemperatureUnits, UnitSystem
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.DataModelBaseClasses.DynamicProperty import DynamicProperty
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Utils.factory_methods import get_empty_dict_union
from ResSimpy.Units.AttributeMappings.DynamicPropertyUnitMapping import OptionsUnits
from ResSimpy.Nexus.NexusKeywords.options_keywords import OPT_KEYWORDS_VALUE_FLOAT, OPT_SINGLE_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.options_keywords import OPT_GLOBAL_METHOD_OVERRIDES_KEYWORDS, OPT_ARRAY_KEYWORDS
from ResSimpy.Utils.general_utilities import check_if_string_is_float


@dataclass(kw_only=True, repr=False)
class NexusOptions(DynamicProperty):
    """Class to hold Nexus options data."""
    file: Optional[NexusFile]
    properties: dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                dict[str, Union[float, pd.DataFrame]]]] = \
        field(default_factory=get_empty_dict_union)
    unit_system: UnitSystem
    __properties_loaded: bool = False  # Used in lazy loading
    __hash__ = DynamicProperty.__hash__

    def __init__(self, file: Optional[NexusFile], model_unit_system: UnitSystem, input_number: int = 1,
                 properties: Optional[dict[str, Union[str, int, float, Enum, list[str], np.ndarray, pd.DataFrame,
                                                      dict[str, Union[float, pd.DataFrame]]]]] = None
                 ) -> None:
        """Initialises the NexusOptions class.

        Args:
            file (Optional[NexusFile]): Nexus options file object
            model_unit_system (UnitSystem): unit system used in the model
            input_number (int): Always set to 1 as there is always one input file
            properties: Dictionary holding Nexus options information. Defaults to empty dictionary
        """
        self.unit_system = model_unit_system
        if properties is not None:
            self.properties = properties
        else:
            self.properties = {}
        if file is None:
            file = NexusFile(location='')
        super().__init__(input_number=input_number, file=file)

    @staticmethod
    def get_keyword_mapping() -> dict[str, tuple[str, type]]:
        """Gets the mapping of nexus keywords to attribute definitions."""
        keywords: dict[str, tuple[str, type]] = {
            'PSTD': ('standard_pressure', float),
            'TSTD': ('standard_temperature', float),
            'RES_TEMP': ('reservoir_temperature', float)
        }
        return keywords

    @property
    def units(self) -> OptionsUnits:
        """Returns the attribute to unit map for the Nexus options class."""
        if not self.__properties_loaded:
            self.load_nexus_options()
        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']
        return OptionsUnits(unit_system=self.unit_system)

    def to_string(self) -> str:
        """Create string with Nexus options data."""
        if not self.__properties_loaded:
            self.load_nexus_options()
        printable_str = ''
        options_dict = self.properties
        for key, value in options_dict.items():
            if key == 'DESC' and isinstance(value, list):
                for desc_line in value:
                    printable_str += 'DESC ' + desc_line + '\n'
            elif isinstance(value, Enum):
                if isinstance(value, UnitSystem) or isinstance(value, TemperatureUnits):
                    printable_str += f'{value.value}\n'
            elif key == 'REGDATA' and isinstance(value, dict):
                for regkey, regvalue in value.items():
                    if isinstance(regvalue, pd.DataFrame):
                        printable_str += f"\nREGDATA {regkey}\n"
                        printable_str += regvalue.to_string(na_rep='', index=False) + '\n'
                        printable_str += 'ENDREGDATA\n'
            elif key not in ['REGDATA', *OPT_GLOBAL_METHOD_OVERRIDES_KEYWORDS]:
                if isinstance(value, np.ndarray):
                    printable_str += f"{key} {' '.join([str(val) for val in value])}\n"
                elif value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        # Work on printing out global method overrides keywords
        found_global_method_overrides_key = False
        for key, value in options_dict.items():
            if key in OPT_GLOBAL_METHOD_OVERRIDES_KEYWORDS:
                if not found_global_method_overrides_key:
                    printable_str += '\nGLOBAL_METHOD_OVERRIDES\n'
                    found_global_method_overrides_key = True
                if value == '':
                    printable_str += f'{key}\n'
                else:
                    printable_str += f'{key} {value}\n'
        if found_global_method_overrides_key:
            printable_str += 'ENDGLOBAL_METHOD_OVERRIDES\n'
        printable_str += '\n'

        return printable_str

    def load_nexus_options(self) -> None:
        """Read Nexus Options file contents and populate NexusOptions object."""
        if self.file is None:
            raise ValueError("NexusOptions file is not set. Please provide a NexusFile object to load Options.")
        file_as_list = self.file.get_flat_list_str_file

        # Check for common input data
        nfo.check_for_and_populate_common_input_data(file_as_list, self.properties)

        # Specify unit system, if provided
        if 'UNIT_SYSTEM' in self.properties.keys() and isinstance(self.properties['UNIT_SYSTEM'], UnitSystem):
            self.unit_system = self.properties['UNIT_SYSTEM']

        # Initialize properties
        # Dictionary to record start and ending indices for regdata tables
        regdata_indices: dict[str, list[int]] = {}
        region_group_name = ''
        start_reading_table = False

        line_indx = 0
        for line in file_as_list:

            # Find OPTIONS key-value pairs, such as PSTD 14.65, or RES_TEMP 200
            potential_keyword = nfo.check_list_tokens(OPT_KEYWORDS_VALUE_FLOAT, line)
            if potential_keyword is not None:
                if potential_keyword in ['PSTD', 'TSTD']:
                    next_value = nfo.get_expected_token_value(potential_keyword, line, file_as_list)
                    if check_if_string_is_float(next_value):
                        self.properties[potential_keyword] = float(next_value)
                    else:  # Float should be expected after next value
                        self.properties[potential_keyword] = float(
                            nfo.get_expected_token_value(next_value, line, file_as_list))
                else:
                    self.properties[potential_keyword] = float(
                        nfo.get_expected_token_value(potential_keyword, line, file_as_list))

            # Find standalone options keywords
            potential_keyword = nfo.check_list_tokens(OPT_SINGLE_KEYWORDS, line)
            if potential_keyword is not None:
                self.properties[potential_keyword] = ''

            # Find option keywords with array values, e.g., BOUNDARY_FLUXIN_SECTORS 1 3 4
            potential_keyword = nfo.check_list_tokens(OPT_ARRAY_KEYWORDS, line)
            if potential_keyword is not None:
                line_elems = line.split('!')[0].split()
                keyword_index = line_elems.index(potential_keyword)
                self.properties[potential_keyword] = np.fromstring(' '.join(line_elems[keyword_index + 1:]), sep=' ')

            # Find global method overrides keywords
            potential_keyword = nfo.check_list_tokens(OPT_GLOBAL_METHOD_OVERRIDES_KEYWORDS, line)
            if potential_keyword is not None:
                next_value = nfo.get_expected_token_value(potential_keyword, line, file_as_list)
                if next_value == 'OFF':
                    self.properties[potential_keyword] = 'OFF'
                else:
                    self.properties[potential_keyword] = ''

            # Find starting and ending indices for regdata table
            if nfo.check_token('REGDATA', line):
                region_group_name = nfo.get_expected_token_value('REGDATA', line, file_as_list)
                regdata_indices[region_group_name] = [line_indx + 1, len(file_as_list)]
                start_reading_table = True
            if start_reading_table and nfo.check_token('ENDREGDATA', line):
                regdata_indices[region_group_name][1] = line_indx
                start_reading_table = False

            line_indx += 1

        # Read in regdata tables
        reg_dfs: dict[str, pd.DataFrame] = {}
        for key in regdata_indices.keys():
            reg_dfs[key] = nfo.read_table_to_df(file_as_list[regdata_indices[key][0]:regdata_indices[key][1]])
        if len(reg_dfs.keys()) > 0:
            self.properties['REGDATA'] = reg_dfs

        self.__properties_loaded = True

    def load_nexus_options_if_not_loaded(self) -> None:
        """Ensures we only load once the options file."""

        if not self.__properties_loaded:
            self.load_nexus_options()
            self.__properties_loaded = True

    def look_up_region_number_by_name(self, region_name: str) -> int:
        """Look up the region number by the region name in the REGDATA.

        Args:
        region_name (str): Name of the region to look up.

        Returns:
        int: The region number if found, otherwise -1.
        """
        region_name = region_name.upper()
        reg_data = self.properties.get('REGDATA', None)
        if reg_data is None or not isinstance(reg_data, dict):
            return -1
        for table in reg_data.values():
            upper_names = table['NAME'].str.upper()
            if upper_names.isin([region_name]).sum() > 0:
                return table[upper_names == region_name]['NUMBER'].to_numpy()[0]
        return -1



"""A class for collecting all NexusWell objects in a NexusSimulator.

Handles adding and removing completions as well as rescheduling wells.
"""
from __future__ import annotations
import warnings
from dataclasses import dataclass, field
from functools import cmp_to_key
from typing import Sequence, Optional, TYPE_CHECKING
from uuid import UUID

import pandas as pd

from ResSimpy.DataModelBaseClasses.DataObjectMixin import DataObjectMixinDictType
from ResSimpy.Enums.HowEnum import OperationEnum
from ResSimpy.Enums.UnitsEnum import UnitSystem
from ResSimpy.Nexus.DataModels.NexusCompletion import NexusCompletion
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.DataModels.NexusWellMod import NexusWellMod
from ResSimpy.Nexus.NexusEnums.DateFormatEnum import DateFormat
from ResSimpy.Nexus.NexusKeywords.wells_keywords import WELLS_KEYWORDS
from ResSimpy.Nexus.nexus_add_new_object_to_file import AddObjectOperations
from ResSimpy.GenericContainerClasses.Wells import Wells
from ResSimpy.Nexus.load_wells import load_wells
import ResSimpy.FileOperations.file_operations as fo
import ResSimpy.Nexus.nexus_file_operations as nfo
from ResSimpy.Utils.invert_nexus_map import attribute_name_to_nexus_keyword
from ResSimpy.Nexus.DataModels.NexusWell import NexusWell

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


@dataclass(kw_only=True)
class NexusWells(Wells):
    """A class for collecting all NexusWell objects in a NexusSimulator.

    Handles adding and removing completions as well as rescheduling wells. This is usually accessed through the
    model.wells property.

    Arguments:
        model (Simulator): NexusSimulator object that has the instance of wells on.
    """
    __model: NexusSimulator = field(compare=False, repr=False)
    __date_format: DateFormat = DateFormat.MM_DD_YYYY
    _wells: list[NexusWell] = field(default_factory=list)

    def __init__(self, model: NexusSimulator) -> None:
        """Initialises the NexusWells class.

        Args:
            model (NexusSimulator): The model object that contains this NexusWells instance.
        """
        self.__model = model
        self.__add_object_operations = AddObjectOperations(NexusCompletion, self.table_header, self.table_footer, model)
        super().__init__(assume_loaded=model.assume_loaded)

    @property
    def model(self) -> NexusSimulator:
        """The model object that contains this NexusWells instance."""
        return self.__model

    @property
    def start_date(self) -> str:
        """Returns start date of the model."""
        return self.__model.start_date

    @property
    def date_format(self) -> DateFormat:
        """Returns date format of the loads wells."""
        # If we haven't loaded in the wells file, load it in case date format is specified there.
        if not self._wells_loaded:
            self._load()
        return self.__date_format

    @property
    def table_header(self) -> str:
        """Returns string representing the start of the table in the wellspec file."""
        return 'WELLSPEC'

    @property
    def table_footer(self) -> str:
        """Returns string representing the end of the table in the file."""
        return ''

    def get_all(self) -> Sequence[NexusWell]:
        """Returns list of all loaded wells in the Nexus file."""
        if not self._wells_loaded:
            self._load()
        return self._wells

    def get(self, well_name: str) -> Optional[NexusWell]:
        """Returns a specific well requested, or None if that well cannot be found."""
        if not self._wells_loaded:
            self._load()

        wells_to_return = filter(lambda x: x.well_name.upper() == well_name.upper(), self._wells)

        return next(wells_to_return, None)

    def get_df(self) -> pd.DataFrame:
        """Returns data frame representing all processed completions data in the Nexus file."""
        # loop through wells and completions to output a table
        if not self._wells_loaded:
            self._load()

        store_dictionaries = []
        for well in self._wells:
            for completion in well.completions:
                completion_props: DataObjectMixinDictType = {
                    'well_name': well.well_name,
                    'units': well.unit_system.name,
                }
                completion_props.update(completion.to_dict())
                store_dictionaries.append(completion_props)
        df_store = pd.DataFrame(store_dictionaries)
        df_store = df_store.dropna(axis=1, how='all')
        return df_store

    def _load(self) -> None:
        if self.__model.model_files.well_files is None:
            raise FileNotFoundError('No wells files found for current model.')
        for method_number, well_file in self.__model.model_files.well_files.items():
            if well_file.location is None:
                warnings.warn(f'Well file location has not been found for {well_file}')
                continue
            wells, date_format = load_wells(nexus_file=well_file, start_date=self.__model.start_date,
                                            default_units=self.__model.default_units, parent_wells_instance=self,
                                            model_date_format=self.__model.date_format)

            self._wells = wells
            self.__date_format = date_format
        self._wells_loaded = True
        # Ensure the newly added wells have additional information populated from the surface file.
        self.model.network.get_load_status()

    def modify(self, well_name: str, completion_properties_list: list[DataObjectMixinDictType],
               how: OperationEnum = OperationEnum.ADD) -> None:
        """Modify the existing wells in memory using a dictionary of properties.

        Args:
        ----
            well_name (str): name of the well to modify
            completion_properties_list (list[InputDict]): a dictionary containing the properties to modify with the \
                attribute as keys and the values as the updated property value. If remove will remove perforation that \
                matches the values in the dictionary.
            how (OperationEnum): operation enum taking the values OperationEnum.ADD, OperationEnum.REMOVE. \
                Specifies how to modify the existing wells perforations.
            remove_all_that_match (bool): If True will remove all wells that partially match the completion_properties\
                provided. If False will remove perforation if only one matches, if several match throws a warning and \
                does not remove them.
            write_to_file (bool): If True writes directly to file. (Currently not in use)
        """
        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well named {well_name} found in simulator')
        for perf in completion_properties_list:
            if how == OperationEnum.ADD:
                try:
                    date = perf.get('date')
                except AttributeError:
                    raise AttributeError(
                        f'No date provided in perf: {perf}, please provide a date to add the perforation at.')
                if date is None:
                    raise AttributeError(
                        f'No date provided in perf: {perf}, please provide a date to add the perforation at.')
                self.add_completion(well_name=well_name, completion_properties=perf)
            elif how == OperationEnum.REMOVE:
                completions_to_remove = well.find_completions(perf)
                well._remove_completions_from_memory(completions_to_remove)
            elif how == OperationEnum.MODIFY:
                raise NotImplementedError('Modify in place not yet available. Please choose one of ADD/REMOVE')
            else:
                raise ValueError('Please select one of the valid OperationEnum values: e.g. OperationEnum.ADD')

    def add_well(self, name: str, units: UnitSystem, completions: Optional[list[NexusCompletion]],
                 wellmods: Optional[list[NexusWellMod]] = None, add_to_file: bool = True) -> NexusWell:
        """Add a new well to the wells collection.

        Args:
            name (str): Name of the well.
            units (UnitSystem): Unit system for the well.
            completions (Optional[list[NexusCompletion]]): List of completions to add to the well.
            wellmods (Optional[list[NexusWellMod]]): List of well modifications to add to the well.
            add_to_file (bool): If True, adds the well to the well file. If False, only adds it to the in-memory list.
        """
        existing_well = self.get(name)
        if existing_well is not None:
            raise ValueError(f'Well with name {name} already exists, cannot add duplicate well names.')
        if add_to_file:
            # the completions will get added later using the add_completion command
            new_completions = []
        else:
            new_completions = completions if completions is not None else []

        new_well = NexusWell(well_name=name, unit_system=units, wellmods=wellmods if wellmods is not None else [],
                             parent_wells_instance=self, completions=new_completions)
        self._wells.append(new_well)

        if not add_to_file:
            return new_well

        # add the well to the well file
        if completions is not None:
            for completion in completions:
                self.add_completion(well_name=name, completion_properties=completion.to_dict(include_nones=False))
        updated_new_well = self.get(name)
        if updated_new_well is None:
            raise ValueError(f'Well {name} could not be added to the wells collection.')
        return updated_new_well

    def add_completion(self, well_name: str, completion_properties: DataObjectMixinDictType,
                       preserve_previous_completions: bool = True, comments: Optional[str] = None) -> None:
        """Adds a completion to an existing wellspec file.

        Args:
            well_name (str): well name to update
            completion_properties (dict[str, float | int | str]): properties of the completion you want to update.
            Must contain date of the completion to be added.
            preserve_previous_completions (bool): if true a new perforation added on a TIME card without a \
            wellspec card for that well will preserve the previous completions from the closest TIME card in addition \
            to the new completion
            comments (Optional[str]): Comments to add to the object line in the file.
        """
        basic_dict: DataObjectMixinDictType = {'name': well_name}
        _, completion_date = self.__add_object_operations.check_name_date(basic_dict | completion_properties)
        well = self.get(well_name)
        well_id = None
        if well is None:
            # ensure it gets added as a new well.
            well = self.add_well(name=well_name, units=self.__model.default_units, completions=None, add_to_file=False)
        else:
            # if the well already exists get the id:
            if len(well.completions) > 0:
                # if the well has completions then get the first completion id
                # this is to ensure that we can find the wellspec file to add the new completion to
                well_id = well.completions[0].id

        # add completion in memory
        new_completion = well._add_completion_to_memory(date=completion_date,
                                                        completion_properties=completion_properties,
                                                        date_format=self.date_format)

        if self.model.model_files.well_files is None:
            raise FileNotFoundError('No well file found, cannot modify ')

        if well_id is not None:
            wellspec_file = self.__add_object_operations.find_which_file_from_id(obj_id=well_id,
                                                                                 file_type_to_search='well_files')
        else:
            # assume to add to the first well file found
            wellspec_file = next(iter(self.model.model_files.well_files.values()))
        # initialise some storage variables
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        new_completion_time_index = -1
        header_index = -1
        headers: list[str] = []
        headers_original: list[str] = []
        additional_headers: list[str] = []
        file_content = wellspec_file.get_flat_list_str_file
        date_found = False
        new_completion_index = len(file_content)
        new_completion_string: list[str] = []
        last_valid_line_index: int = -1
        writing_new_wellspec_table = False
        date_comp = 0
        # if no time cards present in the file just find the name of the well instead
        if not fo.value_in_file('TIME', file_content):
            new_completion_time_index = 0

        for index, line in enumerate(file_content):
            if header_index == -1 and fo.check_token('TIME', line):
                wellspec_date = fo.get_expected_token_value('TIME', line, [line])
                date_comp = self.__model._sim_controls.compare_dates(wellspec_date, completion_date)
                if date_comp == 0:
                    # if we've found the date we're looking for start looking for a wellspec and name card
                    new_completion_time_index = index
                    date_found = True
                    continue
                elif date_comp > 0 and header_index == -1:
                    # if no date is found that is equal and we've found a date that is greater than the specified date
                    # start to compile a wellspec table from scratch and add in the time cards
                    new_completion_index = index
                    header_index = index - 1
                    headers, new_completion_index, new_completion_string, found_completion_at_previous_date = \
                        self.__write_out_existing_wellspec(
                            completion_date, completion_properties, date_found, index,
                            new_completion_index, preserve_previous_completions, well, well_name)
                    writing_new_wellspec_table = True
                    if not found_completion_at_previous_date:
                        break
                else:
                    continue

            if fo.check_token('WELLSPEC', line) and fo.get_token_value('WELLSPEC', line, [line]) == well_name \
                    and new_completion_time_index != -1:
                # get the header of the wellspec table
                header_index, headers, headers_original = self.__add_object_operations.get_and_write_new_header(
                    additional_headers, completion_properties, file_content, index, nexus_mapping, wellspec_file
                )
                continue

            if header_index != -1 and nfo.nexus_token_found(line, WELLS_KEYWORDS) and index > header_index:
                # if we hit the end of the wellspec table for the given well set the index for the new completion
                if last_valid_line_index != -1:
                    new_completion_index = last_valid_line_index + 1
                break

            elif header_index != -1 and index > header_index:
                # check for valid rows + fill extra columns with NA
                line_valid_index = self.__add_object_operations.fill_in_nas(additional_headers, headers_original, index,
                                                                            line, wellspec_file, file_content)
                # set the line to insert the new completion at to be the one after the last valid line
                last_valid_line_index = line_valid_index if line_valid_index > 0 else last_valid_line_index
        # If we haven't found a TIME card after the for loop then we haven't got a valid date so add it at the end
        if date_comp < 0:
            new_completion_index = len(file_content)
            headers, new_completion_index, new_completion_string, found_completion_at_previous_date = \
                self.__write_out_existing_wellspec(
                    completion_date, completion_properties, date_found, new_completion_index,
                    new_completion_index, preserve_previous_completions, well, well_name)
            writing_new_wellspec_table = True

        # construct the new completion and ensure the order of the values is in the same order as the headers
        new_completion_string += [new_completion.to_table_line(headers)]
        new_completion_additional_lines = len(new_completion_string)
        if writing_new_wellspec_table:
            new_completion_string += ['\n']
        # write out to the file_content_as_list
        new_completion_object_ids = {new_completion.id: [new_completion_index + new_completion_additional_lines - 1]}
        wellspec_file.add_to_file_as_list(additional_content=new_completion_string, index=new_completion_index,
                                          additional_objects=new_completion_object_ids, comments=comments)

    def __write_out_existing_wellspec(self, completion_date: str,
                                      completion_properties: DataObjectMixinDictType,
                                      date_found: bool, index: int, new_completion_index: int,
                                      preserve_previous_completions: bool, well: NexusWell, well_name: str) -> \
            tuple[list[str], int, list[str], bool]:
        """Writes out the existing wellspec for a well at a new time stamp, with an additional completion."""
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        completion_table_as_list = ['\n']
        if not date_found:
            completion_table_as_list += ['TIME ' + completion_date + '\n']
        completion_table_as_list += ['WELLSPEC ' + well_name + '\n']
        headers = [k for k, v in nexus_mapping.items() if v[0] in completion_properties]
        if preserve_previous_completions:
            # get all the dates for that well
            date_list = well.dates_of_completions
            previous_dates = [x for x in date_list if
                              self.__model._sim_controls.compare_dates(x, completion_date) < 0]
            if len(previous_dates) == 0:
                # if no dates that are smaller than the completion date then only add the perforation
                # at the current index with a new wellspec card.
                warnings.warn(f'No previous completions found for {well_name} at date: {completion_date}')
                new_completion_index = index
                write_out_headers = [' '.join(headers) + '\n']
                completion_table_as_list += write_out_headers
                return headers, new_completion_index, completion_table_as_list, False

            # get the most recent date that is earlier than the new completion date
            previous_dates = sorted(previous_dates, key=cmp_to_key(self.__model._sim_controls.compare_dates))
            last_date = str(previous_dates[-1])
            completion_to_find: DataObjectMixinDictType = {'date': last_date}
            # find all completions at this date
            previous_completion_list = well.find_completions(completion_to_find)
            if len(previous_completion_list) > 0:
                prev_completion_properties = {k: v for k, v in previous_completion_list[0].to_dict().items() if
                                              v is not None}
                for key in prev_completion_properties:
                    if key == 'date' or key == 'iso_date':
                        continue
                    header_key = attribute_name_to_nexus_keyword(nexus_mapping, key)
                    if header_key not in headers:
                        headers.append(header_key)

            write_out_headers = [' '.join(headers) + '\n']
            completion_table_as_list += write_out_headers
            # run through the existing completions to duplicate the completion at the new time
            for completion in previous_completion_list:
                completion_table_as_list += [completion.to_table_line(headers)]
        else:
            write_out_headers = [' '.join(headers) + '\n']
            completion_table_as_list += write_out_headers
        return headers, new_completion_index, completion_table_as_list, True

    def remove_completion(self, well_name: str,
                          completion_properties: Optional[DataObjectMixinDictType] = None,
                          completion_id: Optional[UUID] = None) -> None:
        """Removes well completion from wellspec file.

        Args:
            well_name(str): Name of the well
            completion_properties(Optional[dict[str, None | float | int | str]]): The completion properties.
            completion_id(Optional[UUID) = None): Completion unique identifier or ID.
        """

        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well found with name: {well_name}')

        if completion_properties is None and completion_id is None:
            raise ValueError('Must provide one of completion_properties dictionary or completion_id.')

        # check for a date:
        if completion_properties is not None:
            completion_date = completion_properties.get('date', 'NO_DATE_PROVIDED')
            if completion_date == 'NO_DATE_PROVIDED':
                raise AttributeError('Completion requires a date. '
                                     'Please provide a date in the completion_properties_list dictionary.')
            if completion_id is None:
                completion_id = well.find_completion(completion_properties).id
        if completion_id is None:
            raise ValueError('No completion found for completion_properties')
        # find which wellspec file we should edit
        wellspec_file = self.__add_object_operations.find_which_file_from_id(obj_id=completion_id,
                                                                             file_type_to_search='well_files')

        # remove from the well object/wells class
        completion_date = well.get_completion_by_id(completion_id).date
        well._remove_completion_from_memory(completion_to_remove=completion_id)

        # drop it from the wellspec file or include file if stored in include file
        if wellspec_file.object_locations is None:
            raise ValueError(f'No object locations specified, cannot find completion id: {completion_id}')
        completion_indices = wellspec_file.object_locations[completion_id]
        if len(completion_indices) > 0:
            for comp_index in completion_indices:
                wellspec_file.remove_from_file_as_list(comp_index, [completion_id])

        # check that we have completions left:
        find_completions_dict: DataObjectMixinDictType = {'date': completion_date}
        remaining_completions = well.find_completions(find_completions_dict)
        if len(remaining_completions) == 0:
            # if there are no more completions remaining for that time stamp then remove the wellspec header!
            self.__remove_wellspec_header(str(completion_date), well_name, wellspec_file)

    def __remove_wellspec_header(self, completion_date: str, well_name: str, wellspec_file: NexusFile) -> None:
        """Removes the wellspec and header if the wellspec table is empty.

        Must first check for whether the well has any remaining completions in the wellspec table.
        """
        nexus_mapping = NexusCompletion.get_keyword_mapping()
        completion_date_found = False
        file_content = wellspec_file.get_flat_list_str_file
        wellspec_index = -1
        header_index = -1
        if not fo.value_in_file('TIME', file_content):
            # if we have no completion date in the file then we have effectively found the right TIME
            completion_date_found = True
        for index, line in enumerate(file_content):
            if fo.check_token('TIME', line) and fo.get_expected_token_value('TIME', line, [line]) == \
                    completion_date:
                completion_date_found = True
            if completion_date_found and fo.check_token('WELLSPEC', line) and \
                    fo.get_token_value('WELLSPEC', line, [line]) == well_name:
                # get the index in the list as string
                wellspec_index = index
                keyword_map = {x: y[0] for x, y in nexus_mapping.items()}
                wellspec_table = file_content[wellspec_index::]
                header_index, _ = nfo.get_table_header(file_as_list=wellspec_table, header_values=keyword_map)
                header_index += wellspec_index
                break
        wellspec_file.remove_from_file_as_list(header_index)
        wellspec_file.remove_from_file_as_list(wellspec_index)

    def modify_completion(self, well_name: str, properties_to_modify: DataObjectMixinDictType,
                          completion_to_change: Optional[DataObjectMixinDictType] = None,
                          completion_id: Optional[UUID] = None,
                          comments: Optional[str] = None) -> None:
        """Modify an existing matching completion, preserves attributes and modifies only additional properties \
        found within the provided properties to modify dictionary.

        Args:
            well_name (str): Name of the well with the completion to be modified.
            properties_to_modify (dict[str, None | float | int | str]): attributes to change to.
            completion_to_change (Optional[dict[str, None | float | int | str]]): properties of the existing completion.
            User must provide enough to uniquely identify the completion.
            completion_id (Optional[UUID]): If provided will match against a known UUID for the completion.
            comments (Optional[str]): Comments to add to the object line in the file.
        """
        well = self.get(well_name)
        if well is None:
            raise ValueError(f'No well found with name: {well_name}')

        if completion_to_change is not None:
            completion = well.find_completion(completion_to_change)
            completion_id = completion.id
        elif completion_id is not None:
            completion = well.get_completion_by_id(completion_id)
        else:
            raise ValueError('Must provide one of completion_to_change dictionary or completion_id')

        # start with the existing properties
        update_completion_properties: DataObjectMixinDictType = \
            {k: v for k, v in completion.to_dict().items() if v is not None}

        update_completion_properties.update(properties_to_modify)

        self.remove_completion(well_name, completion_id=completion_id)
        self.add_completion(well_name, update_completion_properties, preserve_previous_completions=True,
                            comments=comments)



from __future__ import annotations

import os
from typing import Optional, TYPE_CHECKING

import pandas as pd
from ResSimpy.DataModelBaseClasses.GridArrayDefinition import GridArrayDefinition
from ResSimpy.Nexus.DataModels.NexusFile import NexusFile
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_KEYWORDS, STRUCTURED_GRID_KEYWORDS
from ResSimpy.Nexus.NexusKeywords.structured_grid_keywords import GRID_ARRAY_FORMAT_KEYWORDS
import ResSimpy.Nexus.nexus_file_operations as nfo
import ResSimpy.FileOperations.file_operations as fo
from ResSimpy.Utils.general_utilities import check_if_string_is_float

if TYPE_CHECKING:
    from ResSimpy.Nexus.NexusSimulator import NexusSimulator


class StructuredGridOperations:
    def __init__(self, model: NexusSimulator) -> None:
        """Initialises the StructuredGridOperations class.

        Args:
            model (NexusSimulator): Parent model object which the StructuredGridOperations class is associated with.
        """
        self.__model: NexusSimulator = model

    @staticmethod
    def load_token_value_if_present(token: str, modifier: str,
                                    token_property: GridArrayDefinition | dict[str, GridArrayDefinition],
                                    line: str, file_as_list: list[str], line_indx: int,
                                    grid_nexus_file: NexusFile,
                                    original_line_location: int,
                                    file_as_list_with_original_line_numbers: list[tuple[int, str]],
                                    ignore_values: Optional[list[str]] = None,
                                    ) -> None:
        """Gets a token's value if there is one and loads it into the token_property.

        Args:
        ----
            token (str): the token being searched for. e.g. 'PERMX'
            modifier (str): any modifiers applied to the token e.g. 'MULT'
            token_property (GridArrayDefinition | dict[str, GridArrayDefinition]):
                GridArrayDefinition or dict of GridArrayDefinition object to store the modifier and value pair into
            line (str): line to search for the token in
            line_indx (int): index of line in file_as_list
            file_as_list (list[str]): a list of strings containing each line of the file as a new entry
            grid_nexus_file (NexusFile): the NexusFile object containing the grid file.
            original_line_location (int): the line location relative to the expanded file as list with the include
            file paths.
            file_as_list_with_original_line_numbers (list[tuple[int, str]]): a list of tuples containing the line
            number from the expanded file_as_list and the line itself.
            ignore_values (Optional[list[str]]): values to be ignored. Defaults to None.

        Raises:
        ------
            ValueError: raises an error if no numerical value can be found after the supplied token modifier pair
        Returns:
            None: stores the value into token_property supplied instead
        """

        if ignore_values is None:
            ignore_values = []
        region_name: str = ''
        modifier_found = False
        token_found = nfo.check_token(token, line)
        found_modifier = fo.get_previous_value([line])
        token_modifier: str = ''

        if token_found:
            token_next_value = fo.get_token_value(token, line, file_as_list, ignore_values=ignore_values)
        else:
            return
        if token_next_value == found_modifier:
            modifier_found = token_next_value == modifier
            token_modifier = f"{token} {modifier}"
        # If IREGION and the region_group is named
        elif token == 'IREGION' and found_modifier in GRID_ARRAY_FORMAT_KEYWORDS and len(line.strip().split()) == 3:
            region_name = nfo.get_expected_token_value(token, line, file_as_list[line_indx:],
                                                       ignore_values=ignore_values)
            modifier_found = found_modifier == modifier
            token_modifier = f"{token} {region_name} {modifier}"
        if token == 'IREGION' and isinstance(token_property, dict):
            if region_name == '':
                region_name = f"IREG{len(token_property.keys()) + 1}"

        if token_found and modifier_found:
            # If we are loading a multiple, load the two relevant values, otherwise just the next value
            # once we have the region_name, modifier_found, and token_modifier we can break it up in three pieces
            # piece one: make the grid definition array
            # piece two: find out if we have any MOD cards to deal with
            # piece three: make the mod table and add it to the grid array
            relative_line_locations = (
                StructuredGridOperations.__make_grid_def(token, file_as_list, ignore_values, line, line_indx, modifier,
                                                         region_name, token_modifier, token_property, grid_nexus_file,
                                                         original_line_location=original_line_location))
            grid_array_definition = token_property if not isinstance(token_property, dict) else (
                token_property)[region_name]
            relative_line_locations.insert(0, line_indx)

            mod_start_end = StructuredGridOperations.__extract_mod_positions(line_indx, file_as_list)
            if 'VMOD' in mod_start_end:
                vmod_indices = mod_start_end.pop('VMOD')
                StructuredGridOperations.__make_vmod_table(vmod_indices, file_as_list, grid_array_definition,
                                                           grid_nexus_file)
            StructuredGridOperations.__make_mod_table(mod_start_end, file_as_list,
                                                      line, grid_array_definition, token_modifier)
            # update the indices of the loaded gridarraydefinition
            file_line_indices = [file_as_list_with_original_line_numbers[x][0]
                                 for x in relative_line_locations]
            grid_nexus_file.add_object_locations(grid_array_definition.id, file_line_indices)

    @staticmethod
    def __make_grid_def(token: str, file_as_list: list[str], ignore_values: list[str], line: str,
                        line_indx: int, modifier: str, region_name: str, token_modifier: str,
                        token_property: GridArrayDefinition | dict[str, GridArrayDefinition],
                        grid_file: NexusFile, original_line_location: int) -> list[int]:
        """A function that begins the process of populating the grid array.

        Returns:
            list[int]: a list of line locations relative to the file as list.
        """

        object_line_locs_relative_to_file_as_list = []

        if modifier == 'MULT':
            numerical_value, line_loc = fo.get_token_value_with_line_index(
                modifier, line, file_as_list[line_indx:], ignore_values=ignore_values)
            if numerical_value is None or line_loc is None:
                raise ValueError(
                    f'No numerical value found after {token_modifier} keyword in line: {line}')
            object_line_locs_relative_to_file_as_list.append(line_indx + line_loc)
            value_to_multiply, line_loc = fo.get_token_value_with_line_index(
                modifier, line, file_as_list[line_indx:], ignore_values=[numerical_value, *ignore_values])
            if value_to_multiply is None or line_loc is None:
                raise ValueError(
                    f'No value found to multiply after {numerical_value} in line: {line}')
            object_line_locs_relative_to_file_as_list.append(line_indx + line_loc)
            if numerical_value is not None and value_to_multiply is not None:
                if not isinstance(token_property, dict):
                    token_property.modifier = 'MULT'
                    token_property.value = f"{numerical_value} {value_to_multiply}"
                    token_property.name = token
                elif region_name != '':  # IREGION
                    token_property[region_name] = GridArrayDefinition()
                    token_property[region_name].modifier = 'MULT'
                    token_property[region_name].value = f"{numerical_value} {value_to_multiply}"
                    token_property[region_name].name = token
                    token_property[region_name].region_name = region_name
        elif modifier == 'NONE':
            if not isinstance(token_property, dict):
                token_property.modifier = None
                token_property.value = None
                token_property.name = token
            elif region_name != '':
                token_property[region_name] = GridArrayDefinition()
                token_property[region_name].modifier = None
                token_property[region_name].value = None
                token_property[region_name].name = token
                token_property[region_name].region_name = region_name

        else:
            value, line_loc = fo.get_token_value_with_line_index(modifier, line, file_as_list[line_indx:],
                                                                 ignore_values=ignore_values)
            if value is None:
                # Could be 'cut short' by us excluding the rest of a file.
                if not isinstance(token_property, dict):
                    token_property.value = None
                    token_property.modifier = modifier
                    token_property.name = token
                elif region_name != '':  # IREGION
                    token_property[region_name] = GridArrayDefinition()
                    token_property[region_name].modifier = modifier
                    token_property[region_name].value = None
                    token_property[region_name].name = token
                    token_property[region_name].region_name = region_name
            else:
                # Check if VALUE modifier is followed by a numeric string or an INCLUDE file
                line_to_check = file_as_list[line_indx + 1]
                # Check next couple of lines in case next line contains one of ignore_values
                for i in range(line_indx + 1, min(line_indx + 3, len(file_as_list))):
                    if value in file_as_list[i]:
                        line_to_check = file_as_list[i]
                if fo.check_token('INCLUDE', line_to_check):
                    if not isinstance(token_property, dict):
                        token_property.value = value
                        token_property.modifier = modifier
                        token_property.name = token
                        StructuredGridOperations.__add_absolute_path_to_grid_array_definition(
                            grid_array_definition=token_property, line_index_of_include_file=original_line_location,
                            grid_nexus_file=grid_file)
                        object_line_locs_relative_to_file_as_list.append(file_as_list.index(line_to_check))

                    elif region_name != '':  # IREGION
                        region_grid_def = GridArrayDefinition()
                        region_grid_def.modifier = modifier
                        region_grid_def.value = value
                        region_grid_def.name = token
                        region_grid_def.region_name = region_name
                        StructuredGridOperations.__add_absolute_path_to_grid_array_definition(
                            grid_array_definition=region_grid_def,
                            line_index_of_include_file=original_line_location,
                            grid_nexus_file=grid_file)
                        # store it in the dictionary of region grid definitions
                        token_property[region_name] = region_grid_def
                        object_line_locs_relative_to_file_as_list.append(file_as_list.index(line_to_check))

                elif check_if_string_is_float(value[0]) or value[0] == '.':
                    start_indx = line_indx + 1
                    end_indx = len(file_as_list)
                    found_end_value = False
                    for i in range(start_indx, len(file_as_list)):
                        for keyword in STRUCTURED_GRID_KEYWORDS + GRID_ARRAY_KEYWORDS:
                            if nfo.check_token(keyword, file_as_list[i]):
                                end_indx = i
                                found_end_value = True
                                break
                        if found_end_value:
                            break
                    if not isinstance(token_property, dict):
                        token_property.modifier = modifier
                        token_property.name = token
                        token_property.value = '\n'.join([line.strip()
                                                          for line in file_as_list[start_indx:end_indx]]).strip()
                        object_line_locs_relative_to_file_as_list.extend(list(range(start_indx, end_indx)))
                    elif region_name != '':  # IREGION
                        token_property[region_name] = GridArrayDefinition()
                        token_property[region_name].modifier = modifier
                        token_property[region_name].name = token
                        token_property[region_name].region_name = region_name
                        token_property[region_name].value = '\n'.join([line.strip() for line in
                                                                       file_as_list[start_indx:end_indx]]).strip()
                        object_line_locs_relative_to_file_as_list.extend(list(range(start_indx, end_indx)))
                else:  # The grid array keyword is likely inside an include file, presented on previous line
                    if fo.check_token('INCLUDE', file_as_list[line_indx - 1]):
                        object_line_locs_relative_to_file_as_list.append(line_indx - 1)
                        if not isinstance(token_property, dict):
                            token_property.modifier = modifier
                            token_property.value = file_as_list[line_indx - 1].split('INCLUDE')[1].strip()
                            token_property.name = token
                            token_property.keyword_in_include_file = True
                            StructuredGridOperations.__add_absolute_path_to_grid_array_definition(
                                grid_array_definition=token_property, line_index_of_include_file=original_line_location,
                                grid_nexus_file=grid_file
                            )
                        elif region_name != '':  # IREGION
                            token_property[region_name] = GridArrayDefinition()
                            token_property[region_name].modifier = modifier
                            token_property[region_name].name = token
                            token_property[region_name].region_name = region_name
                            token_property[region_name].value \
                                = file_as_list[line_indx - 1].split('INCLUDE')[1].strip()
                            StructuredGridOperations.__add_absolute_path_to_grid_array_definition(
                                grid_array_definition=token_property[region_name],
                                line_index_of_include_file=original_line_location,
                                grid_nexus_file=grid_file
                            )
                            token_property[region_name].keyword_in_include_file = True
                    else:
                        raise ValueError(
                            f'No suitable value found after {token_modifier} keyword in line: {line}')
        return object_line_locs_relative_to_file_as_list

    @staticmethod
    def replace_value(file_as_list: list[str], old_property: GridArrayDefinition, new_property: GridArrayDefinition,
                      token_name: str) -> None:
        """Replace the value and token + modifier with the new values.

        Args:
        ----
            file_as_list (list[str]): a list of strings containing each line of the file as a new entry
            old_property (VariableEntry): property found in the original file to be replaced
            new_property (VariableEntry): new property to replace the old property with
            token_name (str): name of the token being replaced

        Returns:
            None: modifies the file_as_list with the new property
        """

        for line in file_as_list:
            old_token_modifier = f"{token_name} {old_property.modifier}"
            new_token_modifier = f"{token_name} {new_property.modifier}"
            ignore_values = ['INCLUDE'] if old_property.modifier == 'VALUE' else []
            if nfo.check_token(old_token_modifier, line):
                # If we are replacing a mult, replace the first value with a blank
                if old_property.modifier == 'MULT':
                    dummy_value = GridArrayDefinition('MULT', '')
                    fo.get_token_value(old_token_modifier, line, file_as_list, ignore_values=ignore_values,
                                       replace_with=dummy_value)

                fo.get_token_value(old_token_modifier, line, file_as_list, ignore_values=ignore_values,
                                   replace_with=new_property)

                new_line = line.replace(old_token_modifier, new_token_modifier, 1)
                line_index = file_as_list.index(line)
                file_as_list[line_index] = new_line

    @staticmethod
    def append_include_to_grid_file(include_file_location: str, structured_grid_file_path: str) -> None:
        # TODO: change append to be an optional parameter
        """Appends an include file to the end of a grid for adding LGRs.

        Args:
        ----
            include_file_location (str): path to a file to include in the grid
            structured_grid_file_path (str): file path to the structured grid.

        Raises:
        ------
            ValueError: if no structured grid file path is specified in the class instance
        """
        # Get the existing file as a list
        if structured_grid_file_path is None:
            raise ValueError("No file path given or found for structured grid file path. \
                Please update structured grid file path")
        file = nfo.load_file_as_list(structured_grid_file_path)

        file.append(f"\nINCLUDE {include_file_location}\n")
        file.append("TOLPV LGR1 0\n")

        # Save the new file contents
        new_file_str = "".join(file)
        with open(structured_grid_file_path, "w") as text_file:
            text_file.write(new_file_str)

    @staticmethod
    def get_grid_file_as_3d_list(path: str) -> Optional[list]:
        """Converts a grid file to a 3D list.

        Args:
        ----
            path (str): path to a grid file

        Returns:
        -------
            Optional[list[str]]: Returns None if no file is found, returns the grid as a 3d array otherwise
        """
        try:
            with open(path) as f:
                grid_file_list = list(f)
        except FileNotFoundError:
            return None

        sub_lists = []

        new_list_str = ""
        for sub_list in grid_file_list[4:]:
            if sub_list == '\n':
                new_list_split = [x.split("\t") for x in new_list_str.split("\n")]
                new_list_split_cleaned = []
                for x_list in new_list_split:
                    float_list_split = [float(x) for x in x_list if x != ""]
                    new_list_split_cleaned.append(float_list_split)
                sub_lists.append(new_list_split_cleaned)
                new_list_str = ""
            else:
                new_list_str = new_list_str + sub_list
        return sub_lists

    def view_command(self, field: str, previous_lines: int = 3, following_lines: int = 3) -> Optional[str]:
        """Displays how the property is declared in the structured grid file.

        Args:
        ----
            field (str): property as written in the structured grid (e.g. KX)
            previous_lines (int, optional): how many lines to look back from the field searched for. Defaults to 3.
            following_lines (int, optional): how many lines to look forward from the field searched for. Defaults to 3.

        Raises:
        ------
            ValueError: if no structured grid file path is specified in the class instance

        Returns:
        -------
            Optional[str]: the string associated with the supplied property from within the structured grid. \
                If the field is not found in the structured grid returns None.
        """
        structured_grid_dict = self.__model.get_structured_grid_dict()
        structured_grid_path = self.__model.structured_grid_path
        command_token = f"{field.upper()} {structured_grid_dict[field.lower()].modifier}"
        if structured_grid_path is None:
            raise ValueError("No path found for structured grid file path. \
                Please provide a path to the structured grid")
        file_as_list = nfo.load_file_as_list(structured_grid_path)

        for line in file_as_list:
            if nfo.check_token(command_token, line):
                start_index = file_as_list.index(line) - previous_lines \
                    if file_as_list.index(line) - previous_lines > 0 else 0
                end_index = file_as_list.index(line) + following_lines \
                    if file_as_list.index(line) + following_lines < len(file_as_list) else len(file_as_list) - 1

                new_array = file_as_list[start_index: end_index]
                new_array = [x.strip("'") for x in new_array]
                value = "".join(new_array)
                return value
        return None

    @staticmethod
    def __extract_mod_positions(line_indx: int, file_as_list: list[str]) -> dict[str, list[list[int]]]:
        """Finds the positions where MOD exists."""

        mod_start_end: dict[str, list[list[int]]] = {}
        break_flag = False
        keywords_to_stop_on = STRUCTURED_GRID_KEYWORDS + GRID_ARRAY_KEYWORDS
        keywords_to_stop_on.remove('INCLUDE')
        skip_lines = False
        found_end_of_mod_table = False

        for i in range(line_indx + 1, len(file_as_list)):
            line = file_as_list[i]
            # if in a skip line block then skip the line
            if nfo.check_token('SKIP', line):
                skip_lines = True
            if nfo.check_token('NOSKIP', line):
                skip_lines = False
            if skip_lines:
                continue

            one_line_mod_tokens = ['MODX', 'MODY', 'MODZ']
            for token in one_line_mod_tokens:
                if nfo.check_token(token, line):
                    mod_start_end[token] = [[i + 1, i + 2]]
            if nfo.check_token('VMOD', line):
                if mod_start_end.get('VMOD', None) is None:
                    mod_start_end['VMOD'] = []
                mod_start_end['VMOD'].append([i + 1, i + 3])
                for j, vmod_line in enumerate(file_as_list[i + 1::], start=i + 1):
                    # find the include file
                    if nfo.check_token('INCLUDE', vmod_line):
                        mod_start_end['VMOD'][-1][1] = j
                        break
            if nfo.check_token('MOD', line):
                if 'MOD' in mod_start_end.keys():  # Already found a prior mod for this token, append
                    mod_start_end['MOD'].append([i + 1, len(file_as_list)])
                else:
                    mod_start_end['MOD'] = [[i + 1, len(file_as_list)]]
                found_end_of_mod_table = False
                for j in range(i + 1, len(file_as_list)):
                    line_find_end = file_as_list[j]
                    # find the end of the mod table:
                    if nfo.check_token('SKIP', line_find_end):
                        skip_lines = True
                        # terminate reading mods:
                        mod_start_end['MOD'][-1][1] = j
                    if nfo.check_token('NOSKIP', line_find_end):
                        skip_lines = False
                        # restart reading mods
                        mod_start_end['MOD'].append([j + 1, len(file_as_list)])
                    if skip_lines:
                        continue
                    for keyword in keywords_to_stop_on:
                        if nfo.check_token(keyword, line_find_end):
                            mod_start_end['MOD'][-1][1] = j
                            found_end_of_mod_table = True
                            break
                    if found_end_of_mod_table:
                        break
            for keyword in GRID_ARRAY_KEYWORDS:
                if nfo.check_token(keyword, file_as_list[i]):
                    break_flag = True
                    break
            if break_flag:
                break

        return mod_start_end

    @staticmethod
    def __make_mod_table(mod_start_end: dict[str, list[list[int]]], file_as_list: list[str], line: str,
                         grid_array_definition: GridArrayDefinition, token_modifier: str) -> None:
        """A function that creates the mod table."""

        for key in mod_start_end.keys():
            for i in range(len(mod_start_end[key])):
                if mod_start_end[key][i][1] <= mod_start_end[key][i][0]:
                    continue
                file_slice = file_as_list[mod_start_end[key][i][0]:mod_start_end[key][i][1]]
                # exclude lines with INCLUDE keyword
                file_slice = [x for x in file_slice if not nfo.check_token('INCLUDE', x)]
                mod_table = nfo.read_table_to_df(file_slice, noheader=True)
                if len(mod_table.columns) == 7:
                    mod_table.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', '#v']
                elif len(mod_table.columns) == 8:
                    # clean nan's when there is a mix of 7 length and 8 length columns
                    mod_table[7] = mod_table[7].convert_dtypes().astype(str)
                    mod_table[7] = mod_table[7].replace('nan', '')
                    mod_table[7] = mod_table[7].replace('<NA>', '')
                    # Put the last two columns together and drop them to make the #v column
                    mod_table[8] = mod_table[6].astype(str) + mod_table[7].astype(str)
                    mod_table = mod_table.drop([6, 7], axis=1)

                    mod_table.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', '#v']
                else:
                    raise ValueError(
                        f'Unsuitable mod card for {token_modifier} keyword in line: {line}')

                if grid_array_definition.mods is not None:
                    if key in grid_array_definition.mods.keys():
                        orig_mod_tab = grid_array_definition.mods[key].copy()
                        grid_array_definition.mods[key] = \
                            pd.concat([orig_mod_tab, mod_table]).reset_index(drop=True)
                    else:
                        grid_array_definition.mods[key] = mod_table
                else:
                    grid_array_definition.mods = {key: mod_table}

    @staticmethod
    def __add_absolute_path_to_grid_array_definition(grid_array_definition: GridArrayDefinition,
                                                     line_index_of_include_file: int,
                                                     grid_nexus_file: NexusFile,
                                                     ) -> None:
        """Adds the absolute path to the grid array definition.

        Args:
        grid_array_definition (GridArrayDefinition): the grid array definition to add the absolute path to.
        line_index_of_include_file (int): the index of the line in the file containing the include file.
        grid_nexus_file (NexusFile): the NexusFile object representing the top level grid file.
        """
        # cover the trivial case where the path is already absolute
        if grid_array_definition.value is None:
            return
        if os.path.isabs(grid_array_definition.value):
            grid_array_definition.absolute_path = grid_array_definition.value
            return
        if grid_nexus_file is None:
            return

        # create a default to fallback on.
        default_root = os.path.dirname(grid_nexus_file.location)
        absolute_file_path = os.path.join(default_root, grid_array_definition.value)

        line_with_file_uuid = grid_nexus_file.get_flat_list_str_with_file_ids_with_includes
        line, uuid = line_with_file_uuid[line_index_of_include_file]

        # find the include path from within this include file
        if grid_nexus_file.include_objects is None:
            grid_array_definition.absolute_path = absolute_file_path
            return
        matching_includes = [x for x in grid_nexus_file.include_objects if uuid == x.id]
        if len(matching_includes) == 0:
            grid_array_definition.absolute_path = absolute_file_path
            return
        include_file = matching_includes[0]

        absolute_file_path = os.path.join(os.path.dirname(include_file.location), grid_array_definition.value)
        grid_array_definition.absolute_path = absolute_file_path

    @staticmethod
    def __make_vmod_table(vmod_indices: list[list[int]], file_as_list: list[str],
                          grid_array_definition: GridArrayDefinition, grid_nexus_file: NexusFile) -> None:
        """A function that creates the vmod table from a set of line indices.

        Args:
        vmod_indices (list[list[int]]): list of line indices which the VMOD keyword refers to.
        file_as_list (list[str]): a list of strings containing each line of the file as a new entry.
        grid_array_definition (GridArrayDefinition): the grid array definition to add the vmod table to,
        adds the vmod table to the mods attribute dictionary on the grid array definition.
        grid_nexus_file (NexusFile): the NexusFile object representing the top level grid file.
        """
        store_i1, store_i2, store_j1, store_j2, store_k1, store_k2, store_operation, store_include = (
            [], [], [], [], [], [], [], [])

        # assume it is the same as the grid file - this might not work if the vmod is in an include file
        absolute_path_root = os.path.dirname(grid_nexus_file.location)

        for (start_block, end_block) in vmod_indices:
            file_section = file_as_list[start_block:end_block + 1]
            for line in file_section:
                split_line = fo.split_line(line, upper=False)
                if len(split_line) == 7:
                    i1, i2, j1, j2, k1, k2, operation = split_line
                    store_i1.append(int(i1))
                    store_i2.append(int(i2))
                    store_j1.append(int(j1))
                    store_j2.append(int(j2))
                    store_k1.append(int(k1))
                    store_k2.append(int(k2))
                    store_operation.append(operation)
                    continue
                if len(split_line) == 2 and nfo.check_token('INCLUDE', line.upper()):
                    include_file = split_line[1]
                    if not os.path.isabs(include_file):
                        include_file = os.path.join(absolute_path_root, split_line[1])
                    # might need to add absolute path here at some point
                    store_include.append(include_file)
                    continue
        if grid_array_definition.mods is None:
            grid_array_definition.mods = {'VMOD': pd.DataFrame({
                'i1': store_i1, 'i2': store_i2, 'j1': store_j1, 'j2': store_j2, 'k1': store_k1, 'k2': store_k2,
                'operation': store_operation, 'include_file': store_include})}
        else:
            grid_array_definition.mods['VMOD'] = pd.DataFrame({
                'i1': store_i1, 'i2': store_i2, 'j1': store_j1, 'j2': store_j2, 'k1': store_k1, 'k2': store_k2,
                'operation': store_operation, 'include_file': store_include})
